module;

#include <algorithm>
#include <iostream>
#include <limits>
#include <locale>
#include <ranges>
#include <source_location>
#include <sstream>
#include <string>
#include <vector>

export module com.threedbb.ulog;

template<size_t N>
struct StringLiteral {
    constexpr StringLiteral(const char (&str)[N]) {
        std::copy_n(str, N, value);
    }

    char value[N];
};

struct Ostream {
    constexpr Ostream(std::ostream &os) : value(os) {}

    std::ostream &value;
};

template<::StringLiteral LIT, Ostream OUT>
class ULog;

export std::string toUTF8(const std::u16string &str);

export using LDebug = ULog<"DEBUG", std::cout>;
export using LInfo = ULog<"INFO", std::cout>;
export using LWarn = ULog<"WARN", std::cout>;
export using LError = ULog<"ERROR", std::cerr>;

export void LTHROW(const std::string &msg, const std::source_location &loc = std::source_location::current());

std::string asLogFormat(const std::string &level, const std::string &msg, const std::source_location &loc);

template<StringLiteral LIT, Ostream OUT>
class ULog {
private:
    std::ostream &output = OUT.value;
    std::string m_logLevel = LIT.value;
    std::source_location m_loc;
    std::stringstream m_ss;

public:
    explicit ULog(const std::source_location &loc = std::source_location::current())
            : m_loc(loc) {}

    ~ULog() {
        std::string msg;
        while (std::getline(m_ss, msg, '\n')) {
            output << asLogFormat(m_logLevel, msg, m_loc) << "\n";
        }
    }

    template<typename T>
    ULog &operator<<(T const &Xi_val) {
        m_ss << Xi_val;
        return *this;
    }

    ULog &operator<<(const std::u16string &u16) {
        m_ss << toUTF8(u16);
        return *this;
    }

    ULog &operator<<(std::ostream &(*f)(std::ostream &)) {
        if (f == static_cast<decltype(f)>(std::endl)) {
            m_ss << '\n';
        }
        return *this;
    }

    std::ostream &os() {
        return m_ss;
    }
};